/**
 * Copyright (c) 2024 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "drawui.h"
#include "drawshape.h"
#include "assert/assert.h"
#include "ui/textbox.h"
#include "game/time.h"
#include "display/draw/drawtext.h"

void drawUIBox(
  const uint16_t x,
  const uint16_t y,
  const uint16_t width,
  const uint16_t height,
  const uint8_t color,
  const bool_t fill
) {
  assertTrue(x < FRAME_WIDTH, "Box is too far right.");
  assertTrue(y < FRAME_HEIGHT, "Box is too far down.");
  assertTrue(width > 2, "Box is too narrow.");
  assertTrue(height > 2, "Box is too short.");
  assertTrue(x + width <= FRAME_WIDTH, "Box is too wide.");
  assertTrue(y + height <= FRAME_HEIGHT, "Box is too tall.");

  // Draw the top and bottom borders
  for(uint16_t i = 0; i < width; i++) {
    FRAME_BUFFER[y * FRAME_WIDTH + x + i] = '-';
    FRAME_BUFFER[(y + height - 1) * FRAME_WIDTH + x + i] = '-';
    FRAME_COLOR[y * FRAME_WIDTH + x + i] = color;
    FRAME_COLOR[(y + height - 1) * FRAME_WIDTH + x + i] = color;
  }

  // Draw the left and right borders
  for(uint16_t i = 0; i < height; i++) {
    FRAME_BUFFER[(y + i) * FRAME_WIDTH + x] = '|';
    FRAME_BUFFER[(y + i) * FRAME_WIDTH + x + width - 1] = '|';
    FRAME_COLOR[(y + i) * FRAME_WIDTH + x] = color;
    FRAME_COLOR[(y + i) * FRAME_WIDTH + x + width - 1] = color;
  }

  // Draw the corners
  FRAME_BUFFER[y * FRAME_WIDTH + x] = '+';
  FRAME_BUFFER[y * FRAME_WIDTH + x + width - 1] = '+';
  FRAME_BUFFER[(y + height - 1) * FRAME_WIDTH + x] = '+';
  FRAME_BUFFER[(y + height - 1) * FRAME_WIDTH + x + width - 1] = '+';
  FRAME_COLOR[y * FRAME_WIDTH + x] = color;
  FRAME_COLOR[y * FRAME_WIDTH + x + width - 1] = color;
  FRAME_COLOR[(y + height - 1) * FRAME_WIDTH + x] = color;
  FRAME_COLOR[(y + height - 1) * FRAME_WIDTH + x + width - 1] = color;

  // Fill the inside
  if(!fill) return;
  drawBox(x + 1, y + 1, width - 2, height - 2, ' ', COLOR_BLACK);
}

void drawUITextbox() {
  if(!textboxIsOpen()) return;

  // Border
  drawUIBox(
    0, FRAME_HEIGHT - DRAW_UI_TEXTBOX_HEIGHT,
    DRAW_UI_TEXTBOX_WIDTH, DRAW_UI_TEXTBOX_HEIGHT,
    COLOR_MAGENTA, true
  );

  // Title
  drawText(
    TEXTBOX.title,
    -1,
    2, FRAME_HEIGHT - DRAW_UI_TEXTBOX_HEIGHT,
    COLOR_WHITE
  );

  // Text
  drawText(
    TEXTBOX.text,
    TEXTBOX.textIndex,
    1, FRAME_HEIGHT - DRAW_UI_TEXTBOX_HEIGHT + 1,
    COLOR_WHITE
  );

  // Blinking cursor
  memset(&FRAME_BUFFER[DRAW_UI_TEXTBOX_CURSOR_POS], ' ', 3);
  memset(&FRAME_COLOR[DRAW_UI_TEXTBOX_CURSOR_POS], COLOR_WHITE, 3);
  if(TEXTBOX.textIndex < TEXTBOX.textLength) return;
  
  int32_t blink = (int32_t)(TIME.time * DRAW_UI_TEXTBOX_BLINKS_PER_SECOND) % 2;
  FRAME_BUFFER[DRAW_UI_TEXTBOX_CURSOR_POS + 1] = blink ? '>' : ' ';
  FRAME_BUFFER[DRAW_UI_TEXTBOX_CURSOR_POS + 2] = blink ? ' ' : '>';
}