/**
 * Copyright (c) 2024 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "linuxhost.h"
#include "game/game.h"
#include <unistd.h>
#include <termios.h>
#include <time.h>

int32_t LINUX_TERM_HOST_KEY_PRESS;
struct termios origionalTermios;

void terminalModeReset() {
  tcsetattr(0, TCSANOW, &origionalTermios);
}

void terminalModeSetCONIO() {
  struct termios new_termios;

  /* take two copies - one for now, one for later */
  tcgetattr(0, &origionalTermios);
  memcpy(&new_termios, &origionalTermios, sizeof(new_termios));

  /* register cleanup handler, and set the new terminal mode */
  atexit(terminalModeReset);
  cfmakeraw(&new_termios);
  tcsetattr(0, TCSANOW, &new_termios);
}

int32_t keyboardCharacterAvailable() {
  struct timeval tv = { 0L, 0L };
  fd_set fds;
  FD_ZERO(&fds);
  FD_SET(0, &fds);
  return select(1, &fds, NULL, NULL, &tv) > 0;
}

int32_t keyboardCharacterRead() {
  int32_t r;
  uchar_t c;
  if ((r = read(0, &c, sizeof(c))) < 0) {
    return r;
  } else {
    return c;
  }
}

int32_t linuxHostHandoff() {
  gameInit();

  struct timespec spec;
  uint64_t lastTime;
  clock_gettime(CLOCK_REALTIME, &spec);
  lastTime = spec.tv_sec * 1000000000 + spec.tv_nsec;

  while(true) {
    terminalModeSetCONIO();

    while(!keyboardCharacterAvailable()) {
      clock_gettime(CLOCK_REALTIME, &spec);
      uint64_t currentTime = spec.tv_sec * 1000000000 + spec.tv_nsec;
      float_t delta = (currentTime - lastTime) / 1000000000.0f;
      lastTime = currentTime;

      uint8_t result = gameUpdate(delta);
      if(result == GAME_UPDATE_RESULT_EXIT) return 0;
      if(result != GAME_UPDATE_RESULT_CONTINUE) return -1;

      // Clear input.
      LINUX_TERM_HOST_KEY_PRESS = -1;

      // Clear console and return to left top
      printf("\033[2J\033[1;1H");

      // if tick was less than 16ms, sleep for the remaining time
      if(delta < 0.016f) {
        usleep((0.016f - delta) * 1000000);
      }
    }

    LINUX_TERM_HOST_KEY_PRESS = keyboardCharacterRead();
  }

  return 0;
}